""" Generator functions to evaluate trial and test functions

NB: Basis evaluation is only needed for the trial function argument in jacobians, as the
multiplication withthe test function is part of the sum factorization kernel.
"""
from dune.perftool.generation import (backend,
                                      cached,
                                      )
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           basis_functions_per_direction,
                                           name_theta,
                                           quadrature_points_per_direction,
                                           )
from dune.perftool.sumfact.sumfact import (setup_theta,
                                           sum_factorization_kernel,
                                           )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.loopy.buffer import initialize_buffer

from pytools import product

import pymbolic.primitives as p


@cached
def pymbolic_trialfunction(element, restriction, component):
    theta = name_theta()
    rows = quadrature_points_per_direction()
    cols = basis_functions_per_direction()
    a_matrix = AMatrix(theta, rows, cols)
    a_matrices = (a_matrix, a_matrix)

    # Do stage 1
    initialize_buffer("buffer",
                      base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                      num=2
                      )

    insn_dep = setup_theta(element, restriction, component, a_matrices)
    var = sum_factorization_kernel(a_matrices, "buffer", 0, frozenset({insn_dep}))

    return p.Subscript(p.Variable(var), tuple(p.Variable(i) for i in quadrature_inames()))


def lfs_inames(leaf_element):
    return ()


@backend(interface="evaluate_basis")
@cached
def evaluate_basis(leaf_element, name, restriction):
    temporary_variable(name, shape=())
    theta = name_theta()
    quad_inames = quadrature_inames()
    lfs_inames = lfs_inames()
    assert(len(quad_inames) == len(lfs_inames))

    instruction(expression=p.Product(tuple(p.Subscript(p.Variable(theta), (p.Variable(i), p.Variable(j)))
                                           for (i,j) in zip(quad_inames, lfs_inames))
                                     ),
                assignee=p.Variable(name),
                forced_iname_deps=frozenset(quad_inames + lfs_inames),
                forced_iname_deps_is_final=True,
                )
